{-----------------------------------------------------------------------}
{    GPXYSeries jbg                                                }
{                                                                       }
{    30 Jan. 2020@ Control         @                                }
{                                                                       }
{                                                                       }
{                                                                       }
{                                              @copyright   S. Sekine   }
{-----------------------------------------------------------------------}

unit GPXYSeries;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics,  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, System.Generics.Collections,
  System.Math, Vcl.ClipBrd, GPXYGraph,
  Winapi.GDIPAPI, Winapi.GDIPOBJ;

type

  TMarkerShape = (msPlus, msTimes, msSquare, msCircle, msPoint, msStar,
                  msTriangleUp,msTriangleDown, msDiamond);

  TGraphMarker = class(TPersistent)
  private
    FOwnerSeries: TGPXYGraphSeries;
    FShape: TMarkerShape;
    FSize: integer;
    FPen: TPen;
    FBrush: TBrush;
    FVisible: boolean;
    procedure SetShape(const Value: TMarkerShape);
    procedure SetSize(const Value: integer);
    procedure SetPen(const Value: TPen);
    procedure SetBrush(const Value: TBrush);
    procedure SetVisible(const Value: boolean);
    procedure Changed;
    procedure DeviceChanged(Sender: TObject);
  public
    constructor Create(OwnerSeries: TGPXYGraphSeries);
    destructor Destroy;  override;
    procedure Draw(g: TGPGraphics;APen: TGPPen; ABrush: TGPBrush; Center: TPoint);
    procedure Assign(Source: TPersistent); override;
    property OwnerSeries: TGPXYGraphSeries read FOwnerSeries;
  published
    property Pen: TPen
             read FPen write SetPen;
    property Brush: TBrush
             read FBrush write SetBrush;
    property Size: integer
             read FSize write SetSize default 10;
    property Shape: TMarkerShape
             read FShape write SetShape default msDiamond;
    property Visible: boolean
             read FVisible write SetVisible default true;
  end;

  TXYLineSeries = class(TGPXYGraphSeries)
  private
    FLinePen: TPen;
    FDisContinue: boolean;
    FLineThickness: Single;
    procedure SetLineThickness(const Value: Single);
  protected
    procedure SetPolyPolylineArrays(const Points: TGraphPointList;
      const PolyPoints: TGraphIntegerList; StartIndex, Targets: Integer);
    procedure DrawCurve(g: TGPGraphics);
    procedure Draw(g: TGPGraphics); override;
    procedure SetLinePen(const Value: TPen);
    procedure SetDisContinue(const Value: boolean);
    function GetClipCode(var X1,Y1,X2,Y2: double): integer;
    procedure DrawPolyPolyLine(g: TGPGraphics; const Points: TGraphPointList;
      const PolyPoints: TGraphIntegerList); virtual;
    function CreateGPPen(APen: TPen): TGPPen;  override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;  override;
    procedure Assign(Source: TPersistent); override;
    class function GetSeriesTitle: string; override;
  published
    property DisContinue: boolean read FDisContinue write SetDisContinue;
    property LinePen: TPen read FLinePen write SetLinePen;
    property LineThickness: Single read FLineThickness write SetLineThickness;
  end;

  TXYPolygonSeries = class(TXYLineSeries)
  private
    FBrush: TBrush;
    FHatchStyle: THatchStyle;
    FFillForeColor: TColor;
    FFillBackColor: TColor;
    FFillAlpha: Byte;
    procedure SetBrush(const Value: TBrush);
    procedure SetHatchStyle(const Value: THatchStyle);
    procedure SetFillBackColor(const Value: TColor);
    procedure SetFillForeColor(const Value: TColor);
    procedure SetFillAlpha(const Value: Byte);
  protected
    procedure SetPolyPolygonArrays(Points: TGraphPointList;
      PolyPoints: TGraphIntegerList; StartIndex, Targets: Integer);
    procedure DrawPolyPolygon(g: TGPGraphics; const Points: TGraphPointList;
      const PolyPoints: TGraphIntegerList);
    function CreateGPBrush(ABrush: TBrush): TGPBrush;  override;
    procedure DrawPolygon(g: TGPGraphics);
    procedure Draw(g: TGPGraphics); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;  override;
    procedure Assign(Source: TPersistent); override;
    procedure FillSamples(Value: integer); override;
    class function GetSeriesTitle: string; override;
  published
    property Brush: TBrush read FBrush write SetBrush;
    property HatchStyle: THatchStyle read FHatchStyle write SetHatchStyle;
    property FillForeColor: TColor read FFillForeColor write SetFillForeColor;
    property FillBackColor: TColor read FFillBackColor write SetFillBackColor;
    property FillAlpha: Byte read FFillAlpha write SetFillAlpha;
  end;


  TXYMarkerSeries = class(TXYLineSeries)
  private
    FMarker: TGraphMarker;
    FJoined: boolean;
    procedure DrawMarkers(g: TGPGraphics);
  protected
    procedure Draw(g: TGPGraphics); override;
    procedure SetSeriesColor(Value: TColor); override;
    procedure SetMarker(const Value: TGraphMarker);
    procedure SetJoined(const Value: boolean);
    function CreateGPBrush(ABrush: TBrush): TGPBrush;  override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;  override;
    procedure Assign(Source: TPersistent); override;
    class function GetSeriesTitle: string; override;
  published
    property Joined: boolean
             read FJoined write SetJoined;
    property Marker: TGraphMarker
             read FMarker write SetMarker;
  end;

  TXYArrowSeries = class(TXYLineSeries)
  private
    FEndXValues: TGraphValueList;
    FEndYValues: TGraphValueList;
    FArrowSize: integer;
    procedure SetArrowSize(const Value: integer);
    function GetStartXValues: TGraphValueList;
    function GetStartYValues: TGraphValueList;
  protected
    procedure Draw(g: TGPGraphics); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;  override;
    procedure Assign(Source: TPersistent); override;
    class function GetSeriesTitle: string; override;
    function AddSeparator: Integer; override;
    procedure Clear; override;
    procedure Delete(Index: integer); override;
    procedure FillSamples(Value: integer); override;
    function MinXValue: double; override;
    function MaxXValue: double; override;
    function MinYValue: double; override;
    function MaxYValue: double; override;
    function AddXY(X,Y: double): Integer; override;
    function AddArrow(X1,Y1, X2, Y2: double): Integer;
    function AddArrow2(X, Y, Length, Theta: double): Integer;
    property EndXValues: TGraphValueList read FEndXValues;
    property EndYValues: TGraphValueList read FEndYValues;
    property StartXValues: TGraphValueList read GetStartXValues;
    property StartYValues: TGraphValueList read GetStartYValues;
  published
    property ArrowSize: integer read FArrowSize write SetArrowSize default 6;
  end;

  //TXYGraphSeriesEditorClass = class of TXYGraphSeriesEditor;

  TEnumXYGraphSeriesProc =
    procedure(ATitle: string; ASeriesClass: TGPXYGraphSeriesClass) of object;
  procedure RegisterXYGraphSeries(ATitle: string; SeriesClass: TGPXYGraphSeriesClass);
  procedure EnumXYGraphSeries(Proc: TEnumXYGraphSeriesProc);
  function GetXYGraphSeriesClass(ATitle: string): TGPXYGraphSeriesClass;

implementation

uses
  System.UITypes;

var
  XYGraphSeriesList: TStringList;

procedure RegisterXYGraphSeries(ATitle: string; SeriesClass: TGPXYGraphSeriesClass);
begin
  XYGraphSeriesList.AddObject(ATitle, TObject(SeriesClass));
end;

procedure EnumXYGraphSeries(Proc: TEnumXYGraphSeriesProc);
var
  I: Integer;
begin
  with XYGraphSeriesList do
    for I := 0 to Count - 1 do
      Proc(Strings[I], TGPXYGraphSeriesClass(Objects[I]));
end;

function GetXYGraphSeriesClass(ATitle: string): TGPXYGraphSeriesClass;
var
  I: Integer;
begin
  Result := nil;
  with XYGraphSeriesList do
  begin
    for I := 0 to Count - 1 do
      if Strings[I] = ATitle then
      begin
        Result := TGPXYGraphSeriesClass(Objects[I]);
        Break;
      end;
  end;
end;


{TGraphMarker}
constructor TGraphMarker.Create(OwnerSeries: TGPXYGraphSeries);
begin
  inherited Create;
  FOwnerSeries:=OwnerSeries;
  FShape := msDiamond;
  FSize := 10;
  FVisible := true;
  FPen := TPen.Create;
  FPen.Color := clBlack;
  FPen.OnChange := DeviceChanged;
  FBrush := TBrush.Create;
  FBrush.OnChange := DeviceChanged;
end;

destructor TGraphMarker.Destroy;
begin
  FBrush.Free;
  FPen.Free;
  inherited Destroy;
end;

procedure TGraphMarker.Assign(Source: TPersistent);
var
  Marker: TGraphMarker;
begin
  if Source is TGraphMarker then
  begin
    Marker:=TGraphMarker(Source);
    Shape:=Marker.Shape;
    Size:=Marker.Size;
    Pen:=Marker.Pen;
    Brush:=Marker.Brush;
  end;
end;

procedure TGraphMarker.SetShape(const Value: TMarkerShape);
begin
  if FShape<>Value then
  begin
    FShape:=Value;
    Changed;
  end;
end;

procedure TGraphMarker.SetSize(const Value: integer);
begin
  if FSize<>Value then
  begin
    FSize:=Value;
    Changed;
  end;
end;

procedure TGraphMarker.SetPen(const Value: TPen);
begin
  FPen.Assign(Value);
end;

procedure TGraphMarker.SetBrush(const Value: TBrush);
begin
  FBrush.Assign(Value);
end;

procedure TGraphMarker.SetVisible(const Value: boolean);
begin
  if FVisible<>Value then
  begin
    FVisible:=Value;
    Changed;
  end;
end;

procedure TGraphMarker.Draw(g: TGPGraphics;APen: TGPPen; ABrush: TGPBrush; Center: TPoint);
var
  Values: array of integer;
  Points: array of TGPPoint;
  X0, Y0: integer;
begin
  if not FVisible then Exit;
    case FShape of
      msPoint:
        begin
          g.FillEllipse(ABrush, Center.X -1, Center.Y -1 , 2, 2);
          g.DrawEllipse(APen, Center.X -1, Center.Y -1, 2, 2);
        end;
      msPlus:
        begin
          SetLength(Points, 4);
          SetLength(Values, 2);
          Points[0].X := Center.X;
          Points[0].Y := Center.Y - FSize div 2;
          Points[1].X := Center.X;
          Points[1].Y := Points[0].Y +FSize;
          Points[2].X := Center.X - FSize div 2;
          Points[2].Y := Center.Y;
          Points[3].X := Points[2].X + FSize;
          Points[3].Y := Center.Y;
          Values[0]:=2;
          Values[1]:=2;
          g.DrawLine(APen, Points[0], Points[1]);
          g.DrawLine(APen, Points[2], Points[3]);
        end;
      msTimes:
        begin
          SetLength(Points, 4);
          SetLength(Values, 2);
          Points[0].X := Center.X - FSize div 2;
          Points[0].Y := Center.Y - FSize div 2;
          Points[1].X := Points[0].X + FSize;
          Points[1].Y := Points[0].Y + FSize;
          Points[2].X := Points[0].X;
          Points[2].Y := Points[1].Y;
          Points[3].X := Points[1].X;
          Points[3].Y := Points[0].Y;
          Values[0]:=2;
          Values[1]:=2;
          g.DrawLine(APen, Points[0], Points[1]);
          g.DrawLine(APen, Points[2], Points[3]);
        end;
      msStar:
        begin
          SetLength(Points, 8);
          SetLength(Values, 4);
          Points[0].X := Center.X;
          Points[0].Y := Center.Y - FSize div 2;
          Points[1].X := Center.X;
          Points[1].Y := Points[0].Y + FSize;
          Points[2].X := Center.X - FSize div 2;
          Points[2].Y := Center.Y;
          Points[3].X := Points[2].X + FSize;
          Points[3].Y := Center.Y;
          Points[4].X := Points[2].X;
          Points[4].Y := Points[0].Y;
          Points[5].X := Points[3].X;
          Points[5].Y := Points[1].Y;
          Points[6].X := Points[2].X;
          Points[6].Y := Points[1].Y;
          Points[7].X := Points[3].X;
          Points[7].Y := Points[0].Y;
          Values[0]:=2;
          Values[1]:=2;
          Values[2]:=2;
          Values[3]:=2;
          g.DrawLine(APen, Points[0], Points[1]);
          g.DrawLine(APen, Points[2], Points[3]);
          g.DrawLine(APen, Points[4], Points[5]);
          g.DrawLine(APen, Points[6], Points[7]);
        end;
      msTriangleUp:
        begin
          SetLength(Points, 3);
          Points[0].X := Center.X;
          Points[0].Y := Center.Y - FSize div 2;
          Points[1].X := Center.X - FSize div 2;
          Points[1].Y := Center.Y + FSize div 2;
          Points[2].X := Points[1].X + FSize;
          Points[2].Y := Points[1].Y;
          g.FillPolygon(ABrush, PGPPoint(Points), 3);
          g.DrawPolygon(APen, PGPPoint(Points), 3);
        end;
      msTriangleDown:
        begin
          SetLength(Points, 3);
          Points[0].X := Center.X;
          Points[0].Y := Center.Y + FSize div 2;
          Points[1].X := Center.X - FSize div 2;
          Points[1].Y := Center.Y - FSize div 2;
          Points[2].X := Points[1].X + FSize;
          Points[2].Y := Points[1].Y;
          g.FillPolygon(ABrush, PGPPoint(Points), 3);
          g.DrawPolygon(APen, PGPPoint(Points), 3);
        end;
      msSquare:
        begin
          X0 := Center.X - FSize div 2;
          Y0 := Center.Y - FSize div 2;
          g.FillRectangle(ABrush, X0, Y0, FSize, FSize);
          g.DrawRectangle(APen, X0, Y0, FSize, FSize);
        end;
      msDiamond:
        begin
          SetLength(Points, 4);
          Points[0].X := Center.X;
          Points[0].Y := Center.Y - FSize div 2;
          Points[1].X := Center.X - FSize div 2;
          Points[1].Y := Center.Y;

          Points[3].X := Points[1].X + FSize;
          Points[2].Y := Points[0].Y + FSize;
          Points[2].X := Points[3].X - FSize div 2;
          Points[3].Y := Points[2].Y - FSize div 2;
          g.FillPolygon(ABrush, PGPPoint(Points), 4);
          g.DrawPolygon(APen, PGPPoint(Points), 4);
        end;
      msCircle:
        begin
          X0 := Center.X - FSize div 2;
          Y0 := Center.Y - FSize div 2;
          g.FillEllipse(ABrush, X0, Y0, FSize, FSize);
          g.DrawEllipse(APen, X0, Y0, FSize, FSize);
        end;
    end;{of case}
end;

procedure TGraphMarker.Changed;
begin
  if OwnerSeries.ParentGraph <> nil then
    OwnerSeries.ParentGraph.Invalidate;
end;

procedure TGraphMarker.DeviceChanged(Sender: TObject);
begin
  Changed;
end;

{TXYLineSeries}
constructor TXYLineSeries.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDiscontinue := True;
  FLinePen:=TPen.Create;
  //FLinePen.Color := FSeriesColor;
  FLinePen.OnChange:=DeviceChanged;
  FLineThickness := 1.5;
end;

destructor TXYLineSeries.Destroy;
begin
  FLinePen.Free;
  inherited Destroy;
end;

procedure TXYLineSeries.SetLinePen(const Value: TPen);
begin
  FLinePen.Assign(Value);
end;

procedure TXYLineSeries.SetLineThickness(const Value: Single);
begin
  if FLineThickness <> Value then
  begin
    FLineThickness := Value;
    Changed;
  end;
end;

procedure TXYLineSeries.SetDiscontinue(const Value: boolean);
begin
  if FDiscontinue <> Value then
  begin
    FDiscontinue := Value;
    Changed;
  end;
end;

procedure TXYLineSeries.SetPolyPolylineArrays(const Points: TGraphPointList;
  const PolyPoints: TGraphIntegerList; StartIndex, Targets: Integer);
var
  I, ClipCode, Count :integer;
  X1, Y1, X2, Y2, PX, PY: double;
begin
  {2010/01/04C for AddSeparator}
  {Clipping version}
  I := 0;
  while I < Targets do
  begin
    if IsSeparator(I + StartIndex) {and Discontinue} then
    begin
      Inc(I);  continue;
    end;
    X1 := XValues.List[I + StartIndex];
    Y1 := YValues.List[I + StartIndex];
    Count := 0;
    Inc(I);
    while I < Targets do
    begin
      if IsSeparator(I + StartIndex) {and Discontinue} then
      begin
        if (Count > 0) then
        begin
          Points.Add(ParentGraph.GetXYPoint(X1, Y1)); Inc(Count);
          PolyPoints.Add(Count);
        end;
        Inc(I);
        break;
      end;
      X2 := XValues.List[I + StartIndex];
      Y2 := YValues.List[I + StartIndex];
      PX := X2; PY := Y2;
      ClipCode := GetClipCode(X1, Y1, X2, Y2);
      case ClipCode of
        0:
          begin {_2̈OB͓_Qœr؂}
            Points.Add(ParentGraph.GetXYPoint(X1, Y1)); Inc(Count);
            Points.Add(ParentGraph.GetXYPoint(X2, Y2)); Inc(Count);
            PolyPoints.Add(Count);
            Count := 0;
          end;
        1, 2:
          begin {_2͗̈B֑̐}
            Points.Add(ParentGraph.GetXYPoint(X1, Y1)); Inc(Count);
            if I = Targets -1 then {Ō̓_}
            begin
              Points.Add(ParentGraph.GetXYPoint(X2, Y2)); Inc(Count);
              PolyPoints.Add(Count);
            end;
          end;
         {-1 ̏ꍇ͉Ȃ}
      end;{of case}
      X1 := PX; Y1 := PY;
      Inc(I);
    end;{of while}
  end;{of while}
end;

function TXYLineSeries.GetClipCode(var X1,Y1,X2,Y2: double): integer;

  function GetCode(X,Y:double): integer;
  begin
    Result:=0;
    if X < FGraphXMin then Result:=1
    else if X > FGraphXMax then Result:=2;
    if Y < FGraphYMin then Result:=Result or 4
    else if Y > FGraphYMax then Result:=Result or 8;
  end;

var
  Flag: boolean;
  C1, C2: integer;
  dX, dY: double;
begin
  C1:=GetCode(X1,Y1);
  C2:=GetCode(X2,Y2);
  Result:=1;
  if C2>0 then {_2̈O}
    Result:=0;

  Flag := false;
  while (C1 or C2) > 0 do
  begin  {ł̈Oɂꍇ}
    if (C1 and C2)>0  then      {Q_Ԑ͗̈ʂȂ}
    begin
      Result:=-1;
      if (C1 and 8) = 8 then Y1 := FGraphYMax
      else if (C1 and 4) = 4 then Y1 := FGraphYMin;
      if (C1 and 2) = 2 then X1 := FGraphXMax
      else if (C1 and 1) = 1 then X1 := FGraphXMin;
      if (C2 and 8) = 8 then Y2 := FGraphYMax
      else if (C2 and 4) = 4 then Y2 := FGraphYMin;
      if (C2 and 2) = 2 then X2 := FGraphXMax
      else if (C2 and 1) = 1 then X2 := FGraphXMin;
      Exit;
    end;
    dX:=X2-X1; dY:=Y2-Y1;
    if C1=0 then   {_1͗̈ɂꍇ}
    begin
      if (C2 and 8)=8 then begin      { Y2>FGraphYMax }
        X2:=X2-dX*(Y2 - FGraphYMax)/dY;
        Y2:=FGraphYMax;
      end
      else if (C2 and 4)=4 then begin { Y2<FGraphYMin }
        X2:=X2-dX*(Y2 - FGraphYMin)/dY;
        Y2:=FGraphYMin;
      end
      else if (C2 and 2)=2 then begin { X2>FGraphXMax }
        Y2:=Y2-dY*(X2 - FGraphXMax)/dX;
        X2:=FGraphXMax;
      end
      else if (C2 and 1)=1 then begin { X2<FGraphXMin }
        Y2:=Y2-dY*(X2 - FGraphXMin)/dX;
        X2:=FGraphXMin;
      end;
      C2:=GetCode(X2,Y2);
    end
    else begin     {_2̂ق̈ɂꍇ}
      Flag := true;
      if (C1 and 8)=8 then begin      { Y1>FGraphYMax }
        X1:=X1+dX*(FGraphYMax-Y1)/dY;
        Y1:=FGraphYMax;
      end
      else if (C1 and 4)=4 then begin { Y1<FGraphYMin }
        X1:=X1+dX*(FGraphYMin-Y1)/dY;
        Y1:=FGraphYMin;
      end
      else if (C1 and 2)=2 then begin { X1>FGraphXMax }
        Y1:=Y1+dY*(FGraphXMax-X1)/dX;
        X1:=FGraphXMax;
      end
      else if (C1 and 1)=1 then begin { X1<FGraphXMin }
        Y1:=Y1+dY*(FGraphXMin-X1)/dX;
        X1:=FGraphXMin;
      end;
      C1:=GetCode(X1,Y1);
    end;
  end;{ of while }
  if Flag and Discontinue then  {19 Nov 2019L x Discontinue AQߐ\ȂȂB}
    Result := -1;
end;

class function TXYLineSeries.GetSeriesTitle: string;
begin
  Result := '܂'
end;

function TXYLineSeries.CreateGPPen(APen: TPen): TGPPen;
var
  Index: Integer;
begin
  Result := TGPPen.Create(ColorRefToARGB(SeriesColor), FLineThickness);
  if APen.Style = psClear then
    Result.SetColor(MakeColor(0, 0, 0, 0));  //Ȑ
  if APen.Style < psClear then
     Index := Integer(APen.Style);
  Result.SetDashStyle(DashStyle(Index));
end;

procedure TXYLineSeries.DrawCurve(g: TGPGraphics);
var
  Points: TGraphPointList;
  PolyPoints: TGraphIntegerList;
begin
  if ParentGraph = nil then Exit;
  if DataCount=0 then Exit;
  //ParentGraph.AxesFront := False;
  Points := TGraphPointList.Create;
  PolyPoints := TGraphIntegerList.Create;
  try
    SetPolyPolylineArrays(Points, PolyPoints, 0, DataCount);
    FGPPen := CreateGPPen(LinePen);
    try
      DrawPolypolyLine(g, Points, PolyPoints);
    finally
      FreeAndNil(FGPPen);
    end;
  finally
    Points.Free;
    PolyPoints.Free;
  end;
end;

procedure TXYLineSeries.DrawPolyPolyLine(g: TGPGraphics;
  const Points: TGraphPointList; const PolyPoints: TGraphIntegerList);
var
  Path: TGPGraphicsPath;
  I, J, PCount, Index: Integer;
  P1, P2: TGPPoint;
begin
  Path:= TGPGraphicsPath.Create;
  try
    Index := 0;
    for I := 0 to PolyPoints.Count -1  do
    begin
      PCount := PolyPoints[I];
      Path.StartFigure;
      P1 := TGPPoint(Points[Index]);
      for J := 1 to PCount - 1 do
      begin
        P2 := TGPPoint(Points[Index + J]);
        Path.AddLine(P1, P2);
        P1 := P2;
      end;
      Index := Index + PCount;
    end;
    g.DrawPath(FGPPen, Path);
  finally
    FreeAndNil(FGPPen);
    Path.Free;
  end;
end;

procedure TXYLineSeries.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
  if Source is TXYLineSeries then
  begin
    LinePen.Assign(TXYLineSeries(Source).LinePen);
    FDisContinue := TXYLineSeries(Source).DisContinue;
    FLineThickness := TXYLineSeries(Source).LineThickness;
  end;
end;

procedure TXYLineSeries.Draw(g: TGPGraphics);
begin
  DrawCurve(g);
end;

{ TXYPolygonSeries }

constructor TXYPolygonSeries.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBrush := TBrush.Create;
  FBrush.Style := bsSolid;
  FBrush.Color := clRed;
  FBrush.OnChange := DeviceChanged;
  FLinePen.Color := clBlack;
  FHatchStyle := HatchStyleForwardDiagonal;
  FFillForeColor := clRed;
  FFillBackColor := clWhite;
  FFillAlpha := $FF;
end;


{ToDo@HatchStyle ɑΉ邱}
function TXYPolygonSeries.CreateGPBrush(ABrush: TBrush): TGPBrush;
var
  R, G, B: Byte;
  C: ARGB;
begin
  C := ColorRefToARGB(FillForeColor);
  if (ABrush.Style = bsSolid) or (ABrush.Style = bsClear) then
  begin
    R := GetRed(C{ABrush.Color});
    G := GetGreen(C{ABrush.Color});
    B := GetBlue(C{ABrush.Color});
    Result := TGPSolidBrush.Create(MakeColor(FillAlpha, R, G, B));
  end
  else
    {K ABrush.Style  bsCross etc ɐݒ肷邱 }
    Result := TGPHatchBrush.Create(HatchStyle, C, ColorRefToARGB(FillBackColor));
end;

destructor TXYPolygonSeries.Destroy;
begin
  FBrush.Free;
  inherited Destroy;
end;

procedure TXYPolygonSeries.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
  if Source is TXYPolygonSeries then
  begin
    SetBrush(TXYPolygonSeries(Source).Brush);
    FHatchStyle := (TXYPolygonSeries(Source)).HatchStyle;
    FFillAlpha := (TXYPolygonSeries(Source)).FillAlpha;
    FFillBackColor := (TXYPolygonSeries(Source)).FillBackColor;
    FFillForeColor := (TXYPolygonSeries(Source)).FillForeColor;
  end;
end;

procedure TXYPolygonSeries.SetBrush(const Value: TBrush);
begin
  FBrush.Assign(Value);
end;

procedure TXYPolygonSeries.SetFillAlpha(const Value: Byte);
begin
  if FFillAlpha <> Value then
  begin
    FFillAlpha := Value;
    Changed;
  end;
end;

procedure TXYPolygonSeries.SetFillBackColor(const Value: TColor);
begin
  if FFillBackColor <> Value then
  begin
    FFillBackColor := Value;
    Changed;
  end;
end;

procedure TXYPolygonSeries.SetFillForeColor(const Value: TColor);
begin
  if FFillForeColor <> Value then
  begin
    FFillForeColor := Value;
    Changed;
  end;
end;

procedure TXYPolygonSeries.SetHatchStyle(const Value: THatchStyle);
begin
  if FHatchStyle <> Value then
  begin
    FHatchStyle := Value;
    changed;
  end;
end;

procedure TXYPolygonSeries.SetPolyPolygonArrays(Points: TGraphPointList;
  PolyPoints: TGraphIntegerList; StartIndex, Targets: Integer);
var
  I, Count :integer;
  X1, Y1, X2, Y2, PX, PY: double;
begin
  I := 0;
  while I < Targets do
  begin
    if IsSeparator(I + StartIndex)  then
    begin
      Inc(I);  continue;
    end;
    X1 := XValues.List[I + StartIndex];
    Y1 := YValues.List[I + StartIndex];
    Count := 0;
    Inc(I);
    while I < Targets do
    begin
      if IsSeparator(I + StartIndex) then
      begin
        if (Count > 0) then
        begin
          Points.Add(ParentGraph.GetXYPoint(X1, Y1)); Inc(Count);
          PolyPoints.Add(Count);
        end;
        Inc(I);
        break;
      end;
      X2 := XValues.List[I + StartIndex];
      Y2 := YValues.List[I + StartIndex];
      PX := X2; PY := Y2;
      //GetClipCode(X1, Y1, X2, Y2); // Ă͂ȂI
      Points.Add(ParentGraph.GetXYPoint(X1, Y1)); Inc(Count);
      if I = Targets -1 then {Ō̓_}
      begin
        Points.Add(ParentGraph.GetXYPoint(X2, Y2)); Inc(Count);
        PolyPoints.Add(Count);
      end;
      X1 := PX; Y1 := PY;
      Inc(I);
    end;{of while}
  end;{of while}
end;

//Poligon  Path@ĕ`悷B
//2017.09.14@ύX
//ParentGraph.Plot ͈̔͂𒴂ĕ`悵Ȃ悤 g.SetClip@gpB
procedure TXYPolygonSeries.DrawPolygon(g: TGPGraphics);
var
  Points: TGraphPointList;
  PolyPoints: TGraphIntegerList;
  R: TGPRect;
begin
  if ParentGraph = nil then Exit;
  if DataCount = 0 then Exit;
  //ParentGraph.AxesFront := True;
  Points := TGraphPointList.Create;
  PolyPoints := TGraphIntegerList.Create;
  R := MakeRect(ParentGraph.PlotArea);
  g.SetClip(R);
  try
    SetPolyPolygonArrays(Points, PolyPoints, 0, DataCount);
    FGPPen := CreateGPPen(LinePen);
    FGPBrush := CreateGPBrush(Brush);
    try
      DrawPolyPolygon(g, Points, PolyPoints);
    finally
      FreeAndNil(FGPPen);
      FreeAndNil(FGPBrush);
    end;
  finally
    g.ResetClip;
    Points.Free;
    PolyPoints.Free;
  end;
end;

procedure TXYPolygonSeries.DrawPolyPolygon(g: TGPGraphics;
  const Points: TGraphPointList; const PolyPoints: TGraphIntegerList);
var
  Path: TGPGraphicsPath;
  I, J, PCount, Index: Integer;
  P1, P2: TGPPoint;
begin
  Path:= TGPGraphicsPath.Create;
  try
    Index := 0;
    for I := 0 to PolyPoints.Count -1  do
    begin
      PCount := PolyPoints[I];
      Path.StartFigure;
      P1 := TGPPoint(Points[Index]);
      for J := 1 to PCount - 1 do
      begin
        P2 := TGPPoint(Points[Index + J]);
        Path.AddLine(P1 ,P2);
        P1 := P2;
      end;
      Path.CloseFigure;
      Index := Index + PCount;
    end;
    g.DrawPath(FGPPen, Path);
    if Brush.Style <> bsClear then
      g.FillPath(FGPBrush, Path);
  finally
    Path.Free;
  end;
end;

procedure TXYPolygonSeries.FillSamples(Value: integer);
var
  Index, I : integer;
  A, X, Y, dX, dY, Angle: double;
begin
  if Value > 0 then
  begin
    if ParentGraph = nil then
      Error(Name + ' ͐eOtĂ܂');
    ParentGraph.UpdateClipLimits(Self);
    Index := GetIndex;
    Clear;
    Randomize;
    dX := (FGraphXmax - FGraphXmin)/Value;
    dY := (FGraphYmin + FGraphYmax)/Value;
    A := (FGraphYmax - FGraphYmin)/5;
    Angle := 2*Pi/Value;
    for I:=0 to Value  - 1 do
    begin
      X := FGraphXmin + (Index+3) * dX + A * Cos(Angle*I+Random/5);
      Y := FGraphYmin + (Index+3) * dY + A * Sin(Angle*I+Random/5);
      AddXY(X, Y);
    end;
  end;
end;

class function TXYPolygonSeries.GetSeriesTitle: string;
begin
  Result := '|S';
end;

procedure TXYPolygonSeries.Draw(g: TGPGraphics);
begin
  DrawPolygon(g);
end;

{ TXYMarkerSeries }

constructor TXYMarkerSeries.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMarker := TGraphMarker.Create(Self);
  FMarker.Brush.Color := SeriesColor;
end;

function TXYMarkerSeries.CreateGPBrush(ABrush: TBrush): TGPBrush;
var
  R, G, B: Byte;
  C: ARGB;
  Hatch: HatchStyle;
begin
  C := ColorRefToARGB(ABrush.Color);
  if (ABrush.Style = bsSolid) or (ABrush.Style = bsClear) then
  begin
    R := GetRed(C{ABrush.Color});
    G := GetGreen(C{ABrush.Color});
    B := GetBlue(C{ABrush.Color});
    Result := TGPSolidBrush.Create(MakeColor($FF, R, G, B));
  end
  else
  begin
    {K ABrush.Style  bsCross etc ɐݒ肷邱 }
    case ABrush.Style of
      bsHorizontal: Hatch := HatchStyleHorizontal;
      bsVertical:   Hatch := HatchStyleVertical;
      bsFDiagonal:  Hatch := HatchStyleForwardDiagonal;
      bsBDiagonal:  Hatch := HatchStyleBackwardDiagonal;
      bsCross:      Hatch := HatchStyleCross;
      bsDiagCross:  Hatch := HatchStyleDiagonalCross;
      else
        Hatch := HatchStyleCross;
    end;
    Result := TGPHatchBrush.Create(Hatch, C, ColorRefToARGB(clWhite));
  end;
end;

destructor TXYMarkerSeries.Destroy;
begin
  FMarker.Free;
  inherited Destroy;
end;

procedure TXYMarkerSeries.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
  if Source is TXYMarkerSeries then
  begin
    FJoined := TXYMarkerSeries(Source).Joined;
    FMarker.Assign(TXYMarkerSeries(Source).Marker);
  end;
end;

procedure TXYMarkerSeries.DrawMarkers(g: TGPGraphics);
var
  i: integer;
  X, Y:double;
begin
  if Marker.Visible then
  begin
    FGPPen := CreateGPPen(Marker.Pen);
    FGPBrush := CreateGPBrush(Marker.Brush);
    try
      for i:=0 to DataCount-1 do
      begin
        X := XValues.List[i];
        Y := YValues.List[i];
        if (X < FGraphXMin) or (X > FGraphXMax)
          or (Y < FGraphYMin) or (Y > FGraphYMax) then
            Continue;
        Marker.Draw(g, FGPPen, FGPBrush, ParentGraph.GetXYPoint(X, Y));
      end;
    finally
      FreeAndNil(FGPPen);
      FreeAndNil(FGPBrush);
    end;
  end;
end;

class function TXYMarkerSeries.GetSeriesTitle: string;
begin
  Result := '}[J[';
end;

procedure TXYMarkerSeries.Draw(g: TGPGraphics);
begin
  if Joined then
    DrawCurve(g);
  DrawMarkers(g);
end;

procedure TXYMarkerSeries.SetJoined(const Value: boolean);
begin
  if FJoined <> Value then
  begin
    FJoined := Value;
    Changed;
  end;
end;

procedure TXYMarkerSeries.SetMarker(const Value: TGraphMarker);
begin
  if FMarker<>Value then
  begin
    FMarker.Assign(Value);
    Changed;
  end;
end;

procedure TXYMarkerSeries.SetSeriesColor(Value: TColor);
begin
  if FSeriesColor <> Value then
  begin
    FSeriesColor := Value;
    FMarker.Brush.Color := Value;
  end;
end;

{ TXYArrowSeries }
constructor TXYArrowSeries.Create(AOwner: TComponent);
begin
  inherited;
  FArrowSize := 3;
  FEndXValues := TGraphValueList.Create;
  FEndYValues := TGraphValueList.Create;
end;

destructor TXYArrowSeries.Destroy;
begin
  FEndYValues.Free;
  FEndXValues.Free;
  inherited Destroy;
end;

procedure TXYArrowSeries.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
  if Source is TXYArrowSeries then
    FArrowSize := TXYArrowSeries(Source).ArrowSize;
end;

class function TXYArrowSeries.GetSeriesTitle: string;
begin
  Result := '';
end;

function TXYArrowSeries.GetStartXValues: TGraphValueList;
begin
  Result := XValues;
end;

function TXYArrowSeries.GetStartYValues: TGraphValueList;
begin
  Result := YValues;
end;

procedure TXYArrowSeries.SetArrowSize(const Value: integer);
begin
  if (Value >= 0) and (Value <> FArrowSize) then
  begin
    FArrowSize := Value;
    Changed;
  end;
end;

function TXYArrowSeries.AddArrow(X1, Y1, X2, Y2: double): Integer;
begin
  FEndXValues.Add(X2);
  FEndYValues.Add(Y2);
  Result := inherited AddXY(X1, Y1);
end;

function TXYArrowSeries.AddArrow2(X, Y, Length, Theta: double): Integer;
begin
  FEndXValues.Add(X + Length * Cos(Theta));
  FEndYValues.Add(Y + Length * Sin(Theta));
  Result := inherited AddXY(X, Y);
end;

procedure TXYArrowSeries.Clear;
begin
  FEndYValues.Clear;
  FEndXValues.Clear;
  inherited Clear;
end;

procedure TXYArrowSeries.Delete(Index: integer);
begin
  FEndYValues.Delete(Index);
  FEndXValues.Delete(Index);
  inherited Delete(Index);
end;

{20119/11/27 }
{̒[_GDI+̖ŕ`}
procedure TXYArrowSeries.Draw(g: TGPGraphics);
var
  i, ClipCode, dX, dY:integer;
  X1, Y1, X2, Y2: double;
  Points: array[0..1] of TGPPoint;
  EndArrow: TGPAdjustableArrowCap;  //ǉ
begin
  EndArrow := TGPAdjustableArrowCap.create(ArrowSize, ArrowSize, False);
  FGPPen := CreateGPPen(LinePen);
  try
    FGPPen.SetCustomEndCap(EndArrow); //StartCap ɂĂBoJłB
  {Clipping version}
    for i:=0 to DataCount-1 do
    begin
      if IsSeparator(I) then
        Continue;
      X1 := XValues.List[i];
      Y1 := YValues.List[i];
      X2 := EndXValues.List[i];
      Y2 := EndYValues.List[i];
      ClipCode:=GetClipCode(X1,Y1,X2,Y2);
      if ClipCode < 0 then  {̈ʉ߂Ȃ}
        Continue;

    {`}
      Points[0] := TGPPoint(ParentGraph.GetXYPoint(X1,Y1));   //Tail
      Points[1] := TGPPoint(ParentGraph.GetXYPoint(X2,Y2));   //Head
      dX := Points[1].x - Points[0].X;
      if dX < 0 then
        dX := -dX;
      dY := Points[1].y - Points[0].y;
      if dY < 0 then
        dY := -dY;

      if (dX <= 1) and (dY <= 1)then
        g.DrawEllipse(FGPPen, Points[0].x, Points[0].y, 1, 1)
      else
        g.DrawLine(FGPPen, Points[0], Points[1]);
    end;
  finally
    FreeAndNil(EndArrow);
    FreeAndNil(FGPPen);
  end;
end;

function TXYArrowSeries.MaxXValue: double;
begin
  Result := inherited MaxXValue;
  if Result < EndXValues.MaxValue then
    Result := EndXValues.MaxValue;
end;

function TXYArrowSeries.MaxYValue: double;
begin
  Result := inherited MaxYValue;
  if Result < EndYValues.MaxValue then
    Result := EndYValues.MaxValue;
end;

function TXYArrowSeries.MinXValue: double;
begin
  Result := inherited MinXValue;
  if Result > EndXValues.MinValue then
    Result := EndXValues.MinValue;
end;

function TXYArrowSeries.MinYValue: double;
begin
  Result := inherited MinYValue;
  if Result > EndYValues.MinValue then
    Result := EndYValues.MinValue;
end;

procedure TXYArrowSeries.FillSamples(Value: integer);
var
  Index, I : integer;
  A, X, Y, dX, Ymean, Angle: double;
begin
  if Value > 0 then
  begin
    if ParentGraph = nil then
      Error(Name + ' ͐eOtĂ܂');
    ParentGraph.UpdateClipLimits(Self);
    Index := GetIndex;
    Clear;
    Randomize;
    dX := (FGraphXmax - FGraphXmin)/Value;
    A := (FGraphYmax - FGraphYmin)/8;
    Ymean := (FGraphYmin + FGraphYmax)/2;
    for I:=0 to Value  - 1 do
    begin
      X := FGraphXmin + I*dx;
      Y := Ymean + A * Cos(I*Pi/Value - Index * Pi/3)/2 + A * Random/5;
      Angle := (Random {+ Index}) * Pi/2;
      AddArrow(X, Y, X + A * Cos(Angle), Y + A * Sin(Angle));
    end;
  end;
end;

function TXYArrowSeries.AddSeparator: Integer;
begin
  Result := inherited AddSeparator;
  FEndXValues.Add(1);
  FEndYValues.Add(1);
end;

function TXYArrowSeries.AddXY(X, Y: double): Integer;
begin
  Result := AddArrow(X, Y, 0, 0);
end;


initialization
  RegisterClasses([ TXYLineSeries, TXYMarkerSeries,
        TXYArrowSeries, TXYPolygonSeries]);

  XYGraphSeriesList := TStringList.Create;
  RegisterXYGraphSeries('܂', TXYLineSeries);
  RegisterXYGraphSeries('Uz}', TXYMarkerSeries);
  RegisterXYGraphSeries('', TXYArrowSeries);
  RegisterXYGraphSeries('|S', TXYPolygonSeries);


finalization
  XYGraphSeriesList.Free;

end.
